
from functools import partial
import inspect

from .errors import DependentNotFoundError

try:
    import click
except ImportError:
    click = None


class DependencyRegister:
    def __init__(self):
        # Maps dependents to names of resources they require
        self.dependents = {}
        # Maps names of resources to their dependents
        self.resources = {}

    @classmethod
    def _unwrap_func(cls, decorated_func):
        '''
        This unwraps a decorated func, returning the inner wrapped func.

        This may become unnecessary with Python 3.4's inspect.unwrap().
        '''
        if click is not None:
            # Workaround for click.command() decorator not setting
            # __wrapped__
            if isinstance(decorated_func, click.Command):
                return cls._unwrap_func(decorated_func.callback)

        if hasattr(decorated_func, '__wrapped__'):
            # Recursion: unwrap more if needed
            return cls._unwrap_func(decorated_func.__wrapped__)
        else:
            # decorated_func isn't actually decorated, no more
            # unwrapping to do
            return decorated_func

    @classmethod
    def _unwrap_dependent(cls, dependent):
        # Dependent is effectively a class. Classes are registered as is.
        if inspect.isclass(dependent):
            return dependent
        # dependent is some other kind of callable, eg a function
        else:
            return cls._unwrap_func(dependent)

    def _register_dependent(self, dependent, resource_name):
        '''
        Register a mapping of the dependent to resource name.

        After calling, dependency_register.dependents[dependent] should
        contain resource_name.
        '''
        if dependent not in self.dependents:
            self.dependents[dependent] = []
        self.dependents[dependent].insert(0, resource_name)

    def _register_resource_dependency(self, resource_name, dependent):
        if resource_name not in self.resources:
            self.resources[resource_name] = set()
        self.resources[resource_name].add(dependent)

    def register(self, resource_name, dependent=None):
        '''
        Register the given dependent as depending on the "resource"
        named by resource_name.
        '''
        if dependent is None:
            # Give a partial usable as a decorator
            return partial(self.register, resource_name)

        dependent = self._unwrap_dependent(dependent)
        self._register_dependent(dependent, resource_name)
        self._register_resource_dependency(resource_name, dependent)

        # Return dependent to ease use as decorator
        return dependent

    def query_resources(self, dependent):
        dependent = self._unwrap_dependent(dependent)

        if dependent not in self.dependents:
            raise DependentNotFoundError(dependent=dependent)

        return self.dependents[dependent]




from .dependency_register import DependencyRegister
from .resource_provider_register import ResourceProviderRegister
from .resolver import DependencyResolver

class Di:

    def __init__(self, namespace=None):
        self.namespace = namespace
        self.dependencies = DependencyRegister()
        self.providers = ResourceProviderRegister()
        self.resolver = DependencyResolver(
                dependency_register=self.dependencies,
                resource_provider_register=self.providers)

        # For use as a decorator
        self.dependsOn = self.dependencies.register




class FangError(Exception): pass

class DependentNotFoundError(FangError):

    def __init__(self, dependent=None):
        self.dependent = dependent
        if dependent:
            message = (
                    "Couldn't find dependencies registered for {!r}"
                    "".format(dependent))
        else:
            message = (
                    "Couldn't find dependencies registered for the given "
                    "dependent")
        super().__init__(message)

class ProviderAlreadyRegisteredError(FangError):

    def __init__(self, resource_name=None, existing_provider=None):
        self.resource_name = resource_name
        self.existing_provider = existing_provider
        if resource_name and existing_provider:
            message = (
                    'A provider ({provider!r}) has already been '
                    'registered for resource {resource_name!r}'.format(
                        provider=existing_provider,
                        resource_name=resource_name))
        else:
            message = (
                    'A provider has already been registered for the '
                    'resource')
        super().__init__(message)

class ProviderNotFoundError(FangError):

    def __init__(self, resource_name=None):
        self.resource_name = resource_name
        if resource_name:
            message = (
                    "A provider could not be found for resource {!r}"
                    "".format(resource_name))
        else:
            message = (
                    "A provider could not be found for the requested "
                    "resource")
        super().__init__(message)




from .errors import ProviderNotFoundError

# This is effectively what is sometimes termed a "dependency injection
# container".
class DependencyResolver:

    def __init__(
            self,
            dependency_register=None,
            resource_provider_register=None):
        self.dependency_register = dependency_register
        self.resource_provider_register = resource_provider_register

        # Methods delegated to other objects
        self.query_dependents_resources = \
                self.dependency_register.query_resources
        self.resolve = self.resource_provider_register.resolve

    def resolve_all_dependencies(self, dependent):
        return [
                self.resolve(resource_name)
                for resource_name in
                self.query_dependents_resources(dependent)]

    def unpack(self, dependent):
        resources = self.resolve_all_dependencies(dependent)

        # Never return a length-1 list/tuple, to allow easier unpacking
        # eg, avoid need for comma in:
        #   my_one_dep, = my_resolver.unpack_dependencies(my_func)
        if len(resources) == 1:
            return resources[0]
        else:
            return resources

    def are_all_dependencies_met_for(self, dependent):
        for resource_name in self.query_dependents_resources(dependent):
            try:
                self.resolve(resource_name)
            except ProviderNotFoundError as e:
                # TODO: Add error logging here
                return False
        else:
            return True




from functools import partial

from .errors import (
        FangError,
        ProviderAlreadyRegisteredError,
        ProviderNotFoundError)

class ResourceProviderRegister:
    def __init__(self, namespace=None):
        self.namespace = namespace
        # Maps resource names to a provider
        self.resource_providers = {}

    def register(self, resource_name, provider=None, allow_override=False):
        if provider is None:
            # Give a partial usable as a decorator
            return partial(
                    self.register,
                    resource_name, allow_override=allow_override)

        if ((not allow_override) and
                resource_name in self.resource_providers):
            raise ProviderAlreadyRegisteredError(
                    resource_name=resource_name,
                    existing_provider=self.resource_providers[resource_name])

        self.resource_providers[resource_name] = provider

        # Return provider to ease use as decorator
        return provider

    register_callable = register

    # For registering providers which always return the same instance
    def register_instance(self, resource_name, instance=None, **kwargs):
        if instance is None:
            # Give a partial usable as a decorator
            return partial(self.register_instance, resource_name, **kwargs)

        self.register(resource_name, provider=(lambda : instance), **kwargs)
        return instance

    def mass_register(self, resource_names_to_providers, **kwargs):
        for resource_name, provider in resource_names_to_providers.items():
            self.register_instance(resource_name, provider, **kwargs)

    def load(self, other_register, allow_overrides=False):
        if not allow_overrides:
            own_keys = self.resource_providers.keys()
            other_keys = other_register.resource_providers.keys()
            common_keys = own_keys & other_keys
            if common_keys:
                # TODO Add new FangError sub-class?
                raise FangError(
                        'This register already has providers for keys: '
                        '{!r}'.format(common_keys))

        self.resource_providers.update(
                other_register.resource_providers)

    def clear(self):
        self.resource_providers.clear()

    def resolve(self, resource_name):
        if resource_name not in self.resource_providers:
            raise ProviderNotFoundError(resource_name=resource_name)

        return self.resource_providers[resource_name]()
